import os
import sys
import time

from rich.console import Console

from piperider_cli.cli_utils import verify_upload_related_options
from piperider_cli.event import log_event


def run(**kwargs):
    'Profile data source, run assertions, and generate report(s). By default, the raw results and reports are saved in ".piperider/outputs".'

    from piperider_cli.cli_utils import DbtUtil
    from piperider_cli.cli_utils.cloud import CloudConnectorHelper
    from piperider_cli.configuration import FileSystem, is_piperider_workspace_exist
    from piperider_cli.error import DbtProjectNotFoundError, PipeRiderConflictOptionsError
    from piperider_cli.exitcode import EC_WARN_NO_PROFILED_MODULES
    from piperider_cli.generate_report import GenerateReport
    from piperider_cli.guide import Guide
    from piperider_cli.initializer import Initializer
    from piperider_cli.runner import Runner, RunEventPayload

    status = False
    reason = None
    start_time = time.time()
    event_payload = RunEventPayload()
    event_payload.step = 'start'
    try:
        datasource = kwargs.get('datasource')
        table = kwargs.get('table')
        output = kwargs.get('output')
        open_report = kwargs.get('open')
        skip_report = kwargs.get('skip_report')
        dbt_target_path = kwargs.get('dbt_target_path')
        dbt_list = kwargs.get('dbt_list')
        project_name = kwargs.get('project')
        select = kwargs.get('select')
        state = kwargs.get('state')

        enable_upload, enable_share = verify_upload_related_options(**kwargs)

        if project_name is not None:
            os.environ.get('PIPERIDER_API_PROJECT')

        console = Console()
        env_dbt_resources = os.environ.get('PIPERIDER_DBT_RESOURCES')

        # True -> 1, False -> 0
        if sum([True if table else False, dbt_list, env_dbt_resources is not None]) > 1:
            console.print("[bold red]Error:[/bold red] "
                          "['--table', '--dbt-list'] are mutually exclusive")
            sys.exit(1)

        # Search dbt project config files
        dbt_project_dir = kwargs.get('dbt_project_dir')
        no_auto_search = kwargs.get('no_auto_search')
        dbt_project_path = DbtUtil.get_dbt_project_path(dbt_project_dir, no_auto_search, recursive=False)
        dbt_profiles_dir = kwargs.get('dbt_profiles_dir')
        if dbt_project_path:
            working_dir = os.path.dirname(dbt_project_path) if dbt_project_path.endswith('.yml') else dbt_project_path
            FileSystem.set_working_directory(working_dir)
            if dbt_profiles_dir:
                FileSystem.set_dbt_profiles_dir(dbt_profiles_dir)
            # Only run initializer when dbt project path is provided
            Initializer.exec(dbt_project_path=dbt_project_path, dbt_profiles_dir=dbt_profiles_dir, interactive=False)
        elif is_piperider_workspace_exist() is False:
            raise DbtProjectNotFoundError()

        dbt_resources = None
        if select and dbt_list is True:
            raise PipeRiderConflictOptionsError(
                'Cannot use options "--select" with "--dbt-list"',
                hint='Remove "--select" option and use "--dbt-list" instead.'
            )

        if dbt_list:
            dbt_resources = DbtUtil.read_dbt_resources(sys.stdin)
        if env_dbt_resources is not None:
            dbt_resources = DbtUtil.read_dbt_resources(env_dbt_resources)

        ret = Runner.exec(datasource=datasource,
                          table=table,
                          output=output,
                          skip_report=skip_report,
                          dbt_target_path=dbt_target_path,
                          dbt_resources=dbt_resources,
                          dbt_select=select,
                          dbt_state=state,
                          report_dir=kwargs.get('report_dir'),
                          skip_datasource_connection=kwargs.get('skip_datasource'),
                          event_payload=event_payload)
        if ret in (0, EC_WARN_NO_PROFILED_MODULES):

            if not skip_report:
                GenerateReport.exec(None, kwargs.get('report_dir'), output, open_report, open_in_cloud=enable_upload)

            if ret == EC_WARN_NO_PROFILED_MODULES:
                # No module was profiled
                if dbt_list or dbt_resources or select:
                    Guide().show('No resources was profiled. Please check given "--select", "--dbt-list" option or '
                                 'environment variable "PIPERIDER_DBT_RESOURCES" to choose the resources to profile.')
                    ret = 0

            event_payload.step = 'upload'
            if enable_upload:
                ret = CloudConnectorHelper.upload_latest_report(report_dir=kwargs.get('report_dir'),
                                                                debug=kwargs.get('debug'),
                                                                open_report=open_report, enable_share=enable_share,
                                                                project_name=project_name)

        if ret != 0:
            reason = 'error'
            return ret

        status = True
        reason = 'ok'
        event_payload.step = 'done'
        return 0
    except SystemExit as e:
        reason = 'error'
        raise e
    except KeyboardInterrupt as e:
        reason = 'aborted'
        raise e
    except Exception as e:
        reason = 'fatal'
        raise e
    finally:
        end_time = time.time()
        duration = end_time - start_time
        event_payload.status = status
        event_payload.reason = reason
        event_payload.duration = duration
        log_event(event_payload.to_dict(), 'run')



import json
import logging
import os
import re
import sys
import tempfile
import webbrowser
from datetime import datetime

from dateutil import tz
from rich.console import Console
from ruamel import yaml

PIPERIDER_USER_HOME = os.path.expanduser('~/.piperider')
if os.access(os.path.expanduser('~/'), os.W_OK) is False:
    # If we can't create file in user home directory, fallback to use temp folder
    PIPERIDER_USER_HOME = os.path.join(tempfile.gettempdir(), '.piperider')
PIPERIDER_USER_PROFILE = os.path.join(PIPERIDER_USER_HOME, 'profile.yml')


def create_logger(name) -> logging.Logger:
    log_level = logging.WARNING
    if os.environ.get('PIPERIDER_LOG_LEVEL') == 'DEBUG':
        log_level = logging.DEBUG
    if os.environ.get('PIPERIDER_LOG_LEVEL') == 'INFO':
        log_level = logging.INFO

    log = logging.getLogger(name)
    log.setLevel(log_level)
    handler = logging.StreamHandler(sys.stderr)
    handler.setLevel(log_level)
    formatter = logging.Formatter(
        '%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    handler.setFormatter(formatter)
    log.addHandler(handler)
    return log


def is_executed_manually() -> bool:
    # some CI service puts the "CI" var
    if os.environ.get('CI', 'false') == 'true':
        return False

    # For CiecleCI exceptions
    if os.environ.get('CIRCLECI', 'false') == 'true':
        return False

    # if not tty, it probably runs automatically
    if not sys.stdout.isatty():
        return False

    return True


def get_assertion_dir():
    return os.environ.get('PIPERIDER_CUSTOM_ASSERTIONS_PATH', '')


def set_assertion_dir(dir):
    snap_dir = os.path.join(os.path.normpath(dir), '__snapshots__')
    if not os.path.exists(snap_dir):
        os.mkdir(snap_dir)
    os.environ['PIPERIDER_CUSTOM_ASSERTIONS_PATH'] = snap_dir


def get_version():
    version_file = os.path.normpath(os.path.join(os.path.dirname(__file__), 'VERSION'))
    with open(version_file) as fh:
        version = fh.read().strip()
        return version


def get_sentry_dns():
    dns_file = os.path.normpath(os.path.join(os.path.dirname(__file__), 'SENTRY_DNS'))
    with open(dns_file) as f:
        dns = f.read().strip()
        return dns


def get_user_id():
    with open(PIPERIDER_USER_PROFILE, 'r') as f:
        user_profile = yaml.YAML().load(f)
        return user_profile.get('user_id')


def clone_directory(src, dst):
    if sys.version_info >= (3, 8):
        # dirs_exist_ok only available after 3.8
        import shutil
        shutil.copytree(src, dst, dirs_exist_ok=True)
    else:
        from distutils.dir_util import copy_tree
        copy_tree(src, dst)


def datetime_to_str(input, to_tzlocal=False):
    if not isinstance(input, datetime):
        raise ValueError(f"Invalid type: expect 'datetime', got '{type(input).__name__}'")
    if to_tzlocal:
        input = convert_to_tzlocal(input)
    output = input.strftime('%Y-%m-%dT%H:%M:%S.%fZ')
    return output


def str_to_datetime(input: str, to_tzlocal=False):
    output = datetime.strptime(input, '%Y-%m-%dT%H:%M:%S.%fZ')
    if to_tzlocal:
        output = convert_to_tzlocal(output)
    return output


def convert_to_tzlocal(input):
    if not isinstance(input, datetime):
        raise ValueError(f"Invalid type: expect 'datetime', got '{type(input).__name__}'")
    from_zone = tz.tzutc()
    to_zone = tz.tzlocal()
    return input.replace(tzinfo=from_zone).astimezone(to_zone)


__version__ = get_version()


def set_sentry_env():
    if '.dev' in __version__:
        return 'development'
    elif re.match(r'^\d+\.\d+\.\d+\.\d{8}[a|b|rc]?.*$', __version__):
        return 'nightly'
    elif 'a' in __version__:
        return 'alpha'
    elif 'b' in __version__:
        return 'beta'
    elif 'rc' in __version__:
        return 'release-candidate'
    return 'production'


sentry_dns = get_sentry_dns()
sentry_env = set_sentry_env()


def ensure_directory_writable(directory):
    d = os.path.abspath(directory)

    if os.path.exists(d):
        if not os.path.isdir(directory):
            return False
        return os.access(d, os.W_OK)
    else:
        try:
            os.makedirs(directory)
            return True
        except BaseException:
            return False


def raise_exception_when_directory_not_writable(output):
    if output:
        if not ensure_directory_writable(output):
            raise Exception(f'The path "{output}" is not writable')


def safe_load_yaml(file_path):
    try:
        with open(file_path, 'r') as f:
            payload = yaml.safe_load(f)
    except yaml.YAMLError as e:
        print(e)
        return None
    except FileNotFoundError:
        return None
    return payload


def round_trip_load_yaml(file_path):
    with open(file_path, 'r') as f:
        try:
            payload = yaml.round_trip_load(f)
        except yaml.YAMLError as e:
            print(e)
            return None
    return payload


def load_json(file_path):
    with open(file_path, 'r') as f:
        try:
            payload = json.load(f)
        except json.JSONDecodeError as e:
            print(e)
            return None
    return payload


def _init_jinja_env(env):
    def env_var(var, default=None):
        if var is None:
            return var
        return os.getenv(var, default)

    def as_bool(var):
        if var is None:
            return var
        return var.lower() in ('true', 'yes', '1')

    def as_number(var):
        if var is None:
            return var
        if var.isnumeric():
            return int(var)
        return float(var)

    def as_text(var):
        if var is None:
            return var
        return str(var)

    def silent_none(var):
        if var is None:
            return ""
        return var

    env.globals['env_var'] = env_var
    env.filters['as_bool'] = as_bool
    env.filters['as_number'] = as_number
    env.filters['as_text'] = as_text
    env.finalize = silent_none


def load_jinja_template(path: str):
    from jinja2 import Environment, FileSystemLoader

    env = Environment(loader=FileSystemLoader(searchpath=os.path.dirname(path)))
    _init_jinja_env(env)
    template = env.get_template(os.path.basename(path))

    return template


def load_jinja_string_template(value: str):
    from jinja2 import Environment, BaseLoader

    env = Environment(loader=BaseLoader())
    _init_jinja_env(env)
    template = env.from_string(value)

    return template


def open_report_in_browser(report_path='', is_cloud_path=False):
    protocol_prefix = "" if is_cloud_path else "file://"
    try:
        webbrowser.open(f"{protocol_prefix}{report_path}")
    except yaml.YAMLError as e:
        print(e)
        return None


def get_run_json_path(output_dir: str, input=None):
    console = Console()
    if input:
        if not os.path.exists(input):
            console.print(f'[bold red]Error: {input} not found[/bold red]')
            return
        if os.path.isdir(input):
            run_json = os.path.join(input, 'run.json')
        else:
            run_json = input
    else:
        latest = os.path.join(output_dir, 'latest')
        run_json = os.path.join(latest, 'run.json')
        if not os.path.isfile(run_json) and os.path.exists(output_dir):
            latest_report_dir_ctime = 0
            latest_report_dir = ''
            for de in os.scandir(output_dir):
                if not de.is_dir():
                    continue
                if os.path.getctime(de) > latest_report_dir_ctime:
                    latest_report_dir_ctime = os.path.getctime(de)
                    latest_report_dir = de.path
            run_json = os.path.join(output_dir, latest_report_dir, 'run.json')
    return os.path.abspath(run_json)



class RunEventPayload:

    def __init__(self):
        self.status = False
        self.reason = None
        self.step = None
        self.datasource_type = None
        self.skip_datasource = None
        self.tables = 0
        self.columns = []
        self.rows = []
        self.passed_assertions = 0
        self.failed_assertions = 0
        self.passed_dbt_testcases = 0
        self.failed_dbt_testcases = 0
        self.build_in_assertions = 0
        self.custom_assertions = 0
        self.recommended_assertions = 0

    def to_dict(self):
        return self.__dict__


class CompareEventPayload:

    def __init__(self):
        self.status = False
        self.reason = None
        self.step = None
        self.datasource_type = None
        self.skip_datasource = None

    def to_dict(self):
        return self.__dict__



import os

from jsonschema.exceptions import ValidationError
from rich.console import Console
from rich.syntax import Syntax

from piperider_cli.configuration import FileSystem
from piperider_cli.dbtutil import load_dbt_project
from piperider_cli.error import RecipeException
from piperider_cli.recipes import (RecipeConfiguration,
                                   RecipeDbtField, RecipeModel,
                                   RecipePiperiderField, tool)

console = Console()


def _read_dbt_project_file(dbt_project_path):
    if dbt_project_path is None:
        return None
    try:
        dbt_project = load_dbt_project(dbt_project_path)
    except Exception as e:
        console.print(f'[[bold yellow]Skip[/bold yellow]] dbt project: {e}')
        return None
    return dbt_project


def _prepare_dbt_cmds(options: dict, target: bool = False) -> RecipeDbtField:
    dbt_cmds = ['dbt deps']
    if options.get('skip_datasource_connection'):
        dbt_cmds.append('dbt parse')
    else:
        if target:
            select_and_state_options = '--select state:modified+ --state <DBT_STATE_PATH>'
            dbt_cmds.append(f'dbt build {select_and_state_options}')
        else:
            dbt_cmds.append('dbt build')

    return RecipeDbtField(
        {'commands': dbt_cmds}
    )


def _prepare_piperider_cmd(options: dict) -> RecipePiperiderField:
    piperider_cmd = 'piperider run'
    if options.get('skip_datasource_connection'):
        piperider_cmd += ' --skip-datasource'

    return RecipePiperiderField({
        'command': piperider_cmd
    })


def _create_base_recipe(dbt_project_path=None, options: dict = None) -> RecipeModel:
    """
    Create the base recipe
    """
    base = RecipeModel()

    if tool().git_branch() is not None:
        if options.get('base_ref') is not None:
            if tool().git_branch(options.get('base_ref')) is not None:
                base.ref = options.get('base_ref')
            else:
                ex = RecipeException(f"Cannot find specified base ref: {options.get('base_ref')}")
                ex.hint = f"Please check if the specified ref name '{options.get('base_ref')}' is correct."
                raise ex
        else:
            if tool().git_branch('main') is not None:
                base.ref = 'main'
            elif tool().git_branch('master') is not None:
                base.ref = 'master'
            else:
                ex = RecipeException("Cannot find default 'main' or 'master' branch")
                ex.hint = "Please specify the base branch using the '--base-branch' option."
                raise ex

    dbt_project = _read_dbt_project_file(dbt_project_path)
    if dbt_project:
        base.dbt = _prepare_dbt_cmds(options)

    base.piperider = _prepare_piperider_cmd(options)
    return base


def _create_target_recipe(dbt_project_path=None, options: dict = None) -> RecipeModel:
    """
    Create the target recipe
    """
    target = RecipeModel()

    if tool().git_branch() is not None:
        if options.get('target_ref') is not None:
            if tool().git_branch(options.get('target_ref')) is not None:
                target.ref = options.get('target_ref')
            else:
                ex = RecipeException(f"Cannot find specified base ref: {options.get('target_ref')}")
                ex.hint = f"Please check if the specified ref name '{options.get('target_ref')}' is correct."
                raise ex

    dbt_project = _read_dbt_project_file(dbt_project_path)
    if dbt_project:
        target.dbt = _prepare_dbt_cmds(options, target=True)

    target.piperider = _prepare_piperider_cmd(options)
    return target


def generate_default_recipe(overwrite_existing: bool = False,
                            dbt_project_path=None, options: dict = None, interactive: bool = True):
    """
    Generate the default recipe
    """
    recipe_path = FileSystem.DEFAULT_RECIPE_PATH
    if overwrite_existing is True and os.path.exists(recipe_path):
        if interactive is True:
            console.print('[bold green]Piperider default recipe already exist[/bold green]')
        return None
    base = _create_base_recipe(dbt_project_path, options)
    target = _create_target_recipe(dbt_project_path, options)
    recipe = RecipeConfiguration(base=base, target=target)

    try:
        recipe.validate(recipe.__dict__())
    except ValidationError as e:
        console.print(f'[[bold red]Error[/bold red]] Recipe syntax error: {e}')
        return None
    except Exception as e:
        console.print(f'[[bold red]Error[/bold red]] {e}')
        return None

    return recipe


def show_recipe_content(recipe: RecipeConfiguration):
    """
    Display the recipe content
    """

    yaml_output = Syntax(recipe.dump(), 'yaml', theme='monokai', line_numbers=True)
    console.print(yaml_output)
    console.rule('End of Recipe')


if __name__ == '__main__':
    generate_default_recipe(overwrite_existing=True)



import os
from unittest import TestCase, mock

from click.testing import CliRunner

from piperider_cli.cli import init, version, diagnose, run
from piperider_cli.cli_utils import verify_upload_related_options


class TestPipeRiderCli(TestCase):
    def setUp(self) -> None:
        self.cli_runner = CliRunner()
        self.unit_test_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'mock_dbt_project')
        pass

    def test_version_command(self):
        result = self.cli_runner.invoke(version)
        assert result.exit_code == 0

    def test_init_command(self):
        result = self.cli_runner.invoke(init, ["--dbt-project-dir", self.unit_test_dir])
        assert result.exit_code == 0

    @mock.patch('piperider_cli.validator.Validator.diagnose', return_value=True)
    def test_diagnose_command(self, *args):
        result = self.cli_runner.invoke(diagnose, ["--dbt-project-dir", self.unit_test_dir])
        assert result.exit_code == 0

    @mock.patch('piperider_cli.runner.Runner.exec', return_value=3)
    @mock.patch('piperider_cli.generate_report.GenerateReport.exec', return_value=0)
    @mock.patch('piperider_cli.cli_utils.cloud.CloudConnectorHelper.is_login', return_value=True)
    @mock.patch('piperider_cli.cli_utils.cloud.CloudConnectorHelper.upload_latest_report', return_value=0)
    @mock.patch('piperider_cli.runner.RunEventPayload')
    def test_run_command(self, *args):
        # Manually config dbt project and profiles dir
        result = self.cli_runner.invoke(run, [
            "--dbt-project-dir", self.unit_test_dir,
            "--dbt-profiles-dir", self.unit_test_dir,
            "--select", "test",
        ])
        assert result.exit_code == 0

        # Table and dbt-list cannot be used together
        result = self.cli_runner.invoke(run, [
            "--dbt-project-dir", self.unit_test_dir,
            "--table", "test",
            "--dbt-list",
        ])
        assert result.exit_code == 1

        # Select and dbt-list cannot be used together
        result = self.cli_runner.invoke(run, [
            "--dbt-project-dir", self.unit_test_dir,
            "--select", "test",
            "--dbt-list",
        ])
        assert result.exit_code == 1

        # Upload the latest report
        result = self.cli_runner.invoke(run, [
            "--dbt-project-dir", self.unit_test_dir,
            "--upload",
        ])
        assert result.exit_code == 0

        mock_runner_exec = args[-1]
        mock_runner_exec.return_value = 666
        mock_run_event_payload = args[0]
        result = self.cli_runner.invoke(run, [
            "--dbt-project-dir", self.unit_test_dir,
        ])
        assert result.exit_code == 0
        assert mock_run_event_payload.return_value.reason == 'error'

    @mock.patch('piperider_cli.cloud_connector.CloudConnector.is_login', return_value=False)
    @mock.patch('piperider_cli.cloud_connector.CloudConnector.is_auto_upload', return_value=False)
    def test_verify_upload_related_options_without_login(self, *args):
        enable_upload, enable_share = verify_upload_related_options(**{
            'upload': False,
            'share': False,
        })
        assert enable_upload is False and enable_share is False

        enable_upload, enable_share = verify_upload_related_options(**{
            'upload': False,
            'share': True,
        })
        assert enable_upload is False and enable_share is False

        enable_upload, enable_share = verify_upload_related_options(**{
            'upload': True,
            'share': False,
        })
        assert enable_upload is False and enable_share is False

        enable_upload, enable_share = verify_upload_related_options(**{
            'upload': True,
            'share': True,
        })
        assert enable_upload is False and enable_share is False

        enable_upload, enable_share = verify_upload_related_options(**{
            'upload': False,
            'share': True,
            'enable_quick_look_share': True,
        })
        assert enable_upload is True and enable_share is True

    @mock.patch('piperider_cli.cloud_connector.CloudConnector.is_login', return_value=True)
    @mock.patch('piperider_cli.cloud_connector.CloudConnector.is_auto_upload', return_value=False)
    def test_verify_upload_related_options_with_login(self, *args):
        mock_is_auto_upload = args[0]

        enable_upload, enable_share = verify_upload_related_options(**{
            'upload': False,
            'share': False,
        })
        assert enable_upload is False and enable_share is False

        enable_upload, enable_share = verify_upload_related_options(**{
            'upload': False,
            'share': True,
        })
        assert enable_upload is True and enable_share is True

        enable_upload, enable_share = verify_upload_related_options(**{
            'upload': True,
            'share': False,
        })
        assert enable_upload is True and enable_share is False

        mock_is_auto_upload.return_value = True
        enable_upload, enable_share = verify_upload_related_options(**{
            'upload': False,
            'share': False,
        })
        assert enable_upload is True and enable_share is False

        enable_upload, enable_share = verify_upload_related_options(**{
            'upload': False,
            'share': True,
        })
        assert enable_upload is True and enable_share is True
